#include <glib.h>

#include <qmail/qmail.h>

#define ME_FILE      "/control/me"

GQuark v_qmail_error_quark(void)
{
	static GQuark error_quark = 0;

	if (G_UNLIKELY(error_quark == 0))
		error_quark =
			g_quark_from_static_string("v-qmail-error-quark");

	return error_quark;
}

G_CONST_RETURN gchar *v_qmail_me(void)
{
	static gchar *me = NULL;
	GIOChannel *io = NULL;
	GIOStatus status;
	GString *line = NULL;
	GError *error = NULL;
	gchar *tmp;
	gsize end;

	if (me)
		return me;

	tmp = g_strconcat(v_qmail_get_root(), ME_FILE, NULL);
	io = g_io_channel_new_file(tmp, "r", &error);
	g_free(tmp);
	if (!io)
		goto out;

	status = g_io_channel_set_encoding(io, NULL, &error);
	if (status != G_IO_STATUS_NORMAL)
		goto out;

	line = g_string_new(NULL);
	status = g_io_channel_read_line_string(io, line, &end, &error);
	if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_EOF)
		goto out;

	g_string_truncate(line, end);

	me = g_strstrip(g_strdup(line->str));
	if (!me[0]) {
		g_free(me);
		me = NULL;
		goto out;
	}

out:
	if (line)
		g_string_free(line, TRUE);
	if (io)
		g_io_channel_unref(io);

	return me;
}
