#ifndef __V_DOMAIN_DOMAIN_H__
#define __V_DOMAIN_DOMAIN_H__

#include <glib.h>
#include <glib-object.h>

#include <misc/user.h>

G_BEGIN_DECLS

#define V_TYPE_DOMAIN			(v_domain_get_type())
#define V_DOMAIN(o)			(G_TYPE_CHECK_INSTANCE_CAST((o), V_TYPE_DOMAIN, VDomain))
#define V_IS_DOMAIN(o)			(G_TYPE_CHECK_INSTANCE_TYPE((o), V_TYPE_DOMAIN))
#define V_DOMAIN_CLASS(o)		(G_TYPE_CHECK_CLASS_CAST((o), V_TYPE_DOMAIN, VDomainClass))
#define V_IS_DOMAIN_CLASS(o)		(G_TYPE_CHECK_CLASS_TYPE((o), V_TYPE_DOMAIN))
#define V_DOMAIN_GET_CLASS(inst)	(G_TYPE_INSTANCE_GET_CLASS((inst), V_TYPE_DOMAIN, VDomainClass))

#define V_DOMAIN_ERROR			v_domain_error_quark()

typedef enum {
	V_DOMAIN_ERROR_NOENT,
	V_DOMAIN_ERROR_EXIST,
	V_DOMAIN_ERROR_INVALID,
	V_DOMAIN_ERROR_UNSUPP,
	V_DOMAIN_ERROR_FAILED
} VDomainError;

GQuark v_domain_error_quark(void) G_GNUC_CONST;

GType v_domain_get_type(void) G_GNUC_CONST;

VDomain *v_domain_new(const gchar *type, const gchar *domain,
                      gboolean create, GError **error) G_GNUC_MALLOC;

void v_domain_set_auth_method(VDomain *vdomain, GType type);

VDomainEntry *v_domain_get_entry(VDomain *vdomain, const gchar *account,
                                 GError **error);
VAccount *v_domain_get_account(VDomain *vdomain, const gchar *account,
                               GError **error);
VUser *v_domain_get_user(VDomain *vdomain) G_GNUC_CONST;

gboolean v_domain_auth(VDomain *vdomain, const gchar *account,
                       const gchar *pass, GError **error);
VAccount *v_domain_auth_account(VDomain *vdomain, const gchar *account,
                                const gchar *pass, GError **error);

VDomainIter *v_domain_iter_new(VDomain *vdomain) G_GNUC_MALLOC;
VDomainEntry *v_domain_iter_next(VDomainIter *iter, GError **error);
void v_domain_iter_free(VDomainIter *iter);

gboolean v_domain_add_entry(VDomain *vdomain, VDomainEntry *entry,
                            GError **error);
gboolean v_domain_modify_entry(VDomain *vdomain, VDomainEntry *entry,
                               GError **error);
gboolean v_domain_remove_entry(VDomain *vdomain, const gchar *account,
                               GError **error);
void v_domain_abort_changes(VDomain *vdomain);
gboolean v_domain_commit_changes(VDomain *vdomain, GError **error);

void v_domain_set_entry_password(VDomain *domain, VDomainEntry *entry,
                                 const gchar *pass);
gchar *v_domain_build_userdir(VDomain *domain, const gchar *name);

G_END_DECLS

#endif /* __V_DOMAIN_DOMAIN_H__ */
