#include <string.h>

#include <Python.h>
#include <structseq.h>

#include <glib/gconvert.h>
#include <glib/gunicode.h>

#define NO_IMPORT_PYGOBJET
#include <pygobject.h>

#include <misc/vconfig.h>
#include <misc/user.h>
#include <qmail/qmail-maildir.h>
#include <vdomain/vdomain.h>
#include <auth/auth.h>

#define VTOASTER_CHARSET	"iso-8859-15"

static PyTypeObject *_PyGObject_Type;
#define PyGObject_Type (*_PyGObject_Type)

static PyTypeObject *_PyStruct_Passwd_Type;
#define PyStruct_Passwd_Type (*_PyStruct_Passwd_Type)

PyTypeObject G_GNUC_INTERNAL PyVConfig_Type;
PyTypeObject G_GNUC_INTERNAL PyVUser_Type;
PyTypeObject G_GNUC_INTERNAL PyVQMailVirtual_Type;
PyTypeObject G_GNUC_INTERNAL PyVMaildir_Type;
PyTypeObject G_GNUC_INTERNAL PyVAccount_Type;
PyTypeObject G_GNUC_INTERNAL PyVDomain_Type;
PyTypeObject G_GNUC_INTERNAL PyVDomainAuth_Type;
PyTypeObject G_GNUC_INTERNAL PyVDomainEntry_Type;
PyTypeObject G_GNUC_INTERNAL PyVAuth_Type;


#define GETSET_MS(name, klass, convert, var, method, opt) \
static PyObject* \
_wrap_##name##_get_##var(PyGObject *self, void *closure) \
{ \
	klass *entry = convert(self->obj); \
	if (!entry->var) { \
		Py_INCREF(Py_None); \
		return Py_None; \
	} \
	return PyString_FromString(entry->var); \
} \
\
static int \
_wrap_##name##_set_##var(PyGObject *self, PyObject *value, void *closure) \
{ \
	char *str = NULL; \
	if (!opt || value) { \
		str = PyString_AsString(value); \
		if (!str) { \
			PyErr_SetString(PyExc_TypeError, "type mismatch"); \
			return -1; \
		} \
	} \
	name##_set_##method(convert(self->obj), str); \
	return 0; \
}

#define GETSET_MU(name, klass, convert, var, method, opt) \
static PyObject* \
_wrap_##name##_get_##var(PyGObject *self, void *closure) \
{ \
	klass *entry = convert(self->obj); \
	PyObject *ret = NULL; \
	GError *error = NULL; \
	gchar *tmp = NULL; \
	if (entry->var) \
		tmp = g_convert(entry->var, -1, "utf-8", VTOASTER_CHARSET, \
		                NULL, NULL, &error); \
	if (!pyg_error_check(&error) && tmp) \
		ret = PyUnicode_DecodeUTF8(tmp, strlen(tmp), NULL); \
	g_free(tmp); \
	if (!ret) { \
		Py_INCREF(Py_None); \
		return Py_None; \
	} \
	return ret; \
} \
\
static int \
_wrap_##name##_set_##var(PyGObject *self, PyObject *value, void *closure) \
{ \
	GError *error = NULL; \
	char *str = NULL; \
	gchar *tmp = NULL; \
	if (!opt || value) { \
		str = PyString_AsString(PyUnicode_AsUTF8String(value)); \
		if (!str) { \
			PyErr_SetString(PyExc_TypeError, "type mismatch"); \
			return -1; \
		} \
	} \
	tmp = g_convert(str, -1, VTOASTER_CHARSET, "utf-8", \
	                NULL, NULL, &error); \
	if (!pyg_error_check(&error)) {\
		g_free(tmp); \
		return -1; \
	} \
	name##_set_##method(convert(self->obj), tmp); \
	g_free(tmp); \
	return 0; \
}

#define GETSET_X(name, convert, var, type1, type2, type3) \
static PyObject* \
_wrap_##name##_get_##var(PyGObject *self, void *closure) \
{ \
	return Py##type3##_From##type2(convert(self->obj)->var); \
} \
\
static int \
_wrap_##name##_set_##var(PyGObject *self, PyObject *value, void *closure) \
{ \
	if (!value) { \
		PyErr_SetString(PyExc_TypeError, "cannot delete member"); \
		return -1; \
	} \
	if (!Py##type3##_Check(value)) { \
		PyErr_SetString(PyExc_TypeError, "type mismatch"); \
		return -1; \
	} \
	convert(self->obj)->var = Py##type1##_As##type2(value); \
	return 0; \
}



/* ----------- VConfig ----------- */

static int
_wrap_v_config_new(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "filename", NULL };
	char *filename;
	GError *error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s:VConfig.__init__", kwlist, &filename))
		return -1;
	self->obj = (GObject *)v_config_new(filename, &error);
	if (pyg_error_check(&error))
		return -1;

	if (!self->obj) {
		PyErr_SetString(PyExc_RuntimeError, "could not create VConfig object");
		return -1;
	}
	pygobject_register_wrapper((PyObject *)self);
	return 0;
}

static PyObject *
_wrap_v_config_get_default(PyGObject *self)
{
	VConfig *config = v_config_get_default();
	PyObject *ret = pygobject_new((GObject *)config);

	g_object_unref(config);
	return ret;
}

static PyObject *
_wrap_v_config_invalidate_default(PyGObject *self)
{
	v_config_invalidate_default();

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
_wrap_v_config_get_string(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "group", "key", NULL };
	char *group, *key;
	gchar *ret;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "zs:VConfig.get_string", kwlist, &group, &key))
		return NULL;

	ret = v_config_get_string(V_CONFIG(self->obj), group, key);

	if (ret) {
		PyObject *py_ret = PyString_FromString(ret);
		g_free(ret);
		return py_ret;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
_wrap_v_config_get_strv(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "group", "key", NULL };
	char *group, *key;
	gchar **argv;
	PyObject *ret;
	int len = 0, pos;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "zs:VConfig.get_strv", kwlist, &group, &key))
		return NULL;

	argv = v_config_get_strv(V_CONFIG(self->obj), group, key);
	if (argv)
		for(len = 0; argv[len]; len++);

	ret = PyList_New(len);
	for(pos = 0; pos < len; pos++)
		PyList_SetItem(ret, pos, PyString_FromString(argv[pos]));

	return ret;
}

static PyObject *
_wrap_v_config_get_int(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "group", "key", NULL };
	char *group, *key;
	int ret;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "zs:VConfig.get_int", kwlist, &group, &key))
		return NULL;

	ret = v_config_get_int(V_CONFIG(self->obj), group, key);

	return PyInt_FromLong(ret);
}

static PyObject *
_wrap_v_config_get_uint(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "group", "key", NULL };
	char *group, *key;
	unsigned int ret;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "zs:VConfig.get_uint", kwlist, &group, &key))
		return NULL;

	if (v_config_get_uint(V_CONFIG(self->obj), group, key, &ret))
		return PyInt_FromLong(ret);

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
_wrap_v_config_get_ulong(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "group", "key", NULL };
	char *group, *key;
	unsigned long ret;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "zs:VConfig.get_ulong", kwlist, &group, &key))
		return NULL;

	if (v_config_get_ulong(V_CONFIG(self->obj), group, key, &ret))
		return PyLong_FromLong(ret);

	Py_INCREF(Py_None);
	return Py_None;
}

static const PyMethodDef _PyVConfig_methods[] = {
	{ "get_default", (PyCFunction)_wrap_v_config_get_default, METH_NOARGS|METH_STATIC, NULL },
	{ "invalidate_default", (PyCFunction)_wrap_v_config_invalidate_default, METH_NOARGS|METH_STATIC, NULL },
	{ "get_string", (PyCFunction)_wrap_v_config_get_string, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "get_strv", (PyCFunction)_wrap_v_config_get_strv, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "get_int", (PyCFunction)_wrap_v_config_get_int, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "get_uint", (PyCFunction)_wrap_v_config_get_uint, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "get_ulong", (PyCFunction)_wrap_v_config_get_ulong, METH_VARARGS|METH_KEYWORDS, NULL },
	{ NULL, NULL, 0, NULL }
};

PyTypeObject G_GNUC_INTERNAL PyVConfig_Type = {
	PyObject_HEAD_INIT(NULL)
	0,						/* ob_size */
	"vtoaster.Config",				/* tp_name */
	sizeof(PyGObject),				/* tp_basicsize */
	0,						/* tp_itemsize */
	/* methods */
	(destructor)0,					/* tp_dealloc */
	(printfunc)0,					/* tp_print */
	(getattrfunc)0,					/* tp_getattr */
	(setattrfunc)0,					/* tp_setattr */
	(cmpfunc)0,					/* tp_compare */
	(reprfunc)0,					/* tp_repr */
	(PyNumberMethods*)NULL,				/* tp_as_number */
	(PySequenceMethods*)NULL,			/* tp_as_sequence */
	(PyMappingMethods*)NULL,			/* tp_as_mapping */
	(hashfunc)0,					/* tp_hash */
	(ternaryfunc)0,					/* tp_call */
	(reprfunc)0,					/* tp_str */
	(getattrofunc)0,				/* tp_getattro */
	(setattrofunc)0,				/* tp_setattro */
	(PyBufferProcs*)NULL,				/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	/* tp_flags */
	NULL,						/* Documentation string */
	(traverseproc)0,				/* tp_traverse */
	(inquiry)0,					/* tp_clear */
	(richcmpfunc)0,					/* tp_richcompare */
	offsetof(PyGObject, weakreflist),		/* tp_weaklistoffset */
	(getiterfunc)0,					/* tp_iter */
	(iternextfunc)0,				/* tp_iternext */
	(struct PyMethodDef*)_PyVConfig_methods,	/* tp_methods */
	(struct PyMemberDef*)NULL,			/* tp_members */
	(struct PyGetSetDef*)NULL,			/* tp_getset */
	NULL,						/* tp_base */
	NULL,						/* tp_dict */
	(descrgetfunc)0,				/* tp_descr_get */
	(descrsetfunc)0,				/* tp_descr_set */
	offsetof(PyGObject, inst_dict),			/* tp_dictoffset */
	(initproc)_wrap_v_config_new,			/* tp_init */
	(allocfunc)0,					/* tp_alloc */
	(newfunc)0,					/* tp_new */
	(freefunc)0,					/* tp_free */
	(inquiry)0					/* tp_is_gc */
};



/* ----------- VUser ----------- */

static int
_wrap_v_user_new(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "name", NULL };
	char *name;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "z:VUser.__init__", kwlist, &name))
		return -1;
	self->obj = (GObject *)v_user_new(name);

	if (!self->obj) {
		PyErr_SetString(PyExc_RuntimeError, "could not create VUser object");
		return -1;
	}
	pygobject_register_wrapper((PyObject *)self);
	return 0;
}

static PyObject *
_wrap_v_user_become(PyGObject *self)
{
	v_user_become(V_USER(self->obj));

	Py_INCREF(Py_None);
	return Py_None;
}

GETSET_X(v_user, V_USER, user, String, String, String)
GETSET_X(v_user, V_USER, home, String, String, String)
GETSET_X(v_user, V_USER, shell, String, String, String)
GETSET_X(v_user, V_USER, uid, Int, Long, Int)
GETSET_X(v_user, V_USER, gid, Int, Long, Int)

static const PyMethodDef _PyVUser_methods[] = {
	{ "become", (PyCFunction)_wrap_v_user_become, METH_NOARGS, NULL },
	{ NULL, NULL, 0, NULL }
};

static PyGetSetDef _PyVUser_getsets[] = {
	{ "user",	(getter)_wrap_v_user_get_user,		(setter)_wrap_v_user_set_user },
	{ "home",	(getter)_wrap_v_user_get_home,		(setter)_wrap_v_user_set_home },
	{ "shell",	(getter)_wrap_v_user_get_shell,		(setter)_wrap_v_user_set_shell },
	{ "uid",	(getter)_wrap_v_user_get_uid,		(setter)_wrap_v_user_set_uid },
	{ "gid",	(getter)_wrap_v_user_get_gid,		(setter)_wrap_v_user_set_gid },
	{ NULL,		(getter)NULL,				(setter)NULL }
};

PyTypeObject G_GNUC_INTERNAL PyVUser_Type = {
	PyObject_HEAD_INIT(NULL)
	0,						/* ob_size */
	"vtoaster.User",				/* tp_name */
	sizeof(PyGObject),				/* tp_basicsize */
	0,						/* tp_itemsize */
	/* methods */
	(destructor)0,					/* tp_dealloc */
	(printfunc)0,					/* tp_print */
	(getattrfunc)0,					/* tp_getattr */
	(setattrfunc)0,					/* tp_setattr */
	(cmpfunc)0,					/* tp_compare */
	(reprfunc)0,					/* tp_repr */
	(PyNumberMethods*)NULL,				/* tp_as_number */
	(PySequenceMethods*)NULL,			/* tp_as_sequence */
	(PyMappingMethods*)NULL,			/* tp_as_mapping */
	(hashfunc)0,					/* tp_hash */
	(ternaryfunc)0,					/* tp_call */
	(reprfunc)0,					/* tp_str */
	(getattrofunc)0,				/* tp_getattro */
	(setattrofunc)0,				/* tp_setattro */
	(PyBufferProcs*)NULL,				/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	/* tp_flags */
	NULL,						/* Documentation string */
	(traverseproc)0,				/* tp_traverse */
	(inquiry)0,					/* tp_clear */
	(richcmpfunc)0,					/* tp_richcompare */
	offsetof(PyGObject, weakreflist),		/* tp_weaklistoffset */
	(getiterfunc)0,					/* tp_iter */
	(iternextfunc)0,				/* tp_iternext */
	(struct PyMethodDef*)_PyVUser_methods,		/* tp_methods */
	(struct PyMemberDef*)NULL,			/* tp_members */
	(struct PyGetSetDef*)_PyVUser_getsets,		/* tp_getset */
	NULL,						/* tp_base */
	NULL,						/* tp_dict */
	(descrgetfunc)0,				/* tp_descr_get */
	(descrsetfunc)0,				/* tp_descr_set */
	offsetof(PyGObject, inst_dict),			/* tp_dictoffset */
	(initproc)_wrap_v_user_new,			/* tp_init */
	(allocfunc)0,					/* tp_alloc */
	(newfunc)0,					/* tp_new */
	(freefunc)0,					/* tp_free */
	(inquiry)0					/* tp_is_gc */
};



/* ----------- VQMailVirtual ----------- */

static int
_wrap_v_qmail_virtual_new(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	GError *error = NULL;

	self->obj = (GObject *)v_qmail_virtual_new(&error);
	if (pyg_error_check(&error))
		return -1;

	if (!self->obj) {
		PyErr_SetString(PyExc_RuntimeError, "could not create VQMailVirtual object");
		return -1;
	}
	pygobject_register_wrapper((PyObject *)self);
	return 0;
}

static PyObject *
_wrap_v_qmail_virtual_get_default(PyGObject *self)
{
	VQMailVirtual *qmail_virtual = v_qmail_virtual_get_default();
	PyObject *ret = pygobject_new((GObject *)qmail_virtual);

	g_object_unref(qmail_virtual);
	return ret;
}

static PyObject *
_wrap_v_qmail_virtual_invalidate_default(PyGObject *self)
{
	v_qmail_virtual_invalidate_default();

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
_wrap_v_qmail_virtual_changed(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	return PyBool_FromLong(
		v_qmail_virtual_changed(V_QMAIL_VIRTUAL(self->obj)));
}

static PyObject *
_wrap_v_qmail_virtual_lookup(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "domain", NULL };
	char *domain;
	const gchar *ret;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s:VQMailVirtual.lookup", kwlist, &domain))
		return NULL;

	ret = v_qmail_virtual_lookup(V_QMAIL_VIRTUAL(self->obj), domain);

	if (ret) {
		PyObject *py_ret = PyString_FromString(ret);
		return py_ret;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static const PyMethodDef _PyVQMailVirtual_methods[] = {
	{ "get_default", (PyCFunction)_wrap_v_qmail_virtual_get_default, METH_NOARGS|METH_STATIC, NULL },
	{ "invalidate_default", (PyCFunction)_wrap_v_qmail_virtual_invalidate_default, METH_NOARGS|METH_STATIC, NULL },
	{ "changed", (PyCFunction)_wrap_v_qmail_virtual_changed, METH_NOARGS, NULL },
	{ "lookup", (PyCFunction)_wrap_v_qmail_virtual_lookup, METH_VARARGS|METH_KEYWORDS, NULL },
	{ NULL, NULL, 0, NULL }
};

PyTypeObject G_GNUC_INTERNAL PyVQMailVirtual_Type = {
	PyObject_HEAD_INIT(NULL)
	0,						/* ob_size */
	"vtoaster.Virtual",				/* tp_name */
	sizeof(PyGObject),				/* tp_basicsize */
	0,						/* tp_itemsize */
	/* methods */
	(destructor)0,					/* tp_dealloc */
	(printfunc)0,					/* tp_print */
	(getattrfunc)0,					/* tp_getattr */
	(setattrfunc)0,					/* tp_setattr */
	(cmpfunc)0,					/* tp_compare */
	(reprfunc)0,					/* tp_repr */
	(PyNumberMethods*)NULL,				/* tp_as_number */
	(PySequenceMethods*)NULL,			/* tp_as_sequence */
	(PyMappingMethods*)NULL,			/* tp_as_mapping */
	(hashfunc)0,					/* tp_hash */
	(ternaryfunc)0,					/* tp_call */
	(reprfunc)0,					/* tp_str */
	(getattrofunc)0,				/* tp_getattro */
	(setattrofunc)0,				/* tp_setattro */
	(PyBufferProcs*)NULL,				/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	/* tp_flags */
	NULL,						/* Documentation string */
	(traverseproc)0,				/* tp_traverse */
	(inquiry)0,					/* tp_clear */
	(richcmpfunc)0,					/* tp_richcompare */
	offsetof(PyGObject, weakreflist),		/* tp_weaklistoffset */
	(getiterfunc)0,					/* tp_iter */
	(iternextfunc)0,				/* tp_iternext */
	(struct PyMethodDef*)_PyVQMailVirtual_methods,	/* tp_methods */
	(struct PyMemberDef*)NULL,			/* tp_members */
	(struct PyGetSetDef*)NULL,			/* tp_getset */
	NULL,						/* tp_base */
	NULL,						/* tp_dict */
	(descrgetfunc)0,				/* tp_descr_get */
	(descrsetfunc)0,				/* tp_descr_set */
	offsetof(PyGObject, inst_dict),			/* tp_dictoffset */
	(initproc)_wrap_v_qmail_virtual_new,			/* tp_init */
	(allocfunc)0,					/* tp_alloc */
	(newfunc)0,					/* tp_new */
	(freefunc)0,					/* tp_free */
	(inquiry)0					/* tp_is_gc */
};



/* ----------- VMaildir ----------- */

static int
_wrap_v_maildir_new(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "name", "create", NULL };
	char *name;
	int create = 0;
	GError *error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|i:VMaildir.__init__", kwlist, &name, &create))
		return -1;
	if (create)
		self->obj = (GObject *)v_maildir_new_create(name, &error);
	else
		self->obj = (GObject *)v_maildir_new(name, &error);
	if (pyg_error_check(&error))
		return -1;

	if (!self->obj) {
		PyErr_SetString(PyExc_RuntimeError, "could not create VMaildir object");
		return -1;
	}
	pygobject_register_wrapper((PyObject *)self);
	return 0;
}

static PyObject *
_wrap_v_maildir_get_base(PyGObject *self)
{
	gchar *ret;

	ret = v_maildir_get_base(V_MAILDIR(self->obj));

	if (ret) {
		PyObject *py_ret = PyString_FromString(ret);
		g_free(ret);
		return py_ret;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
_wrap_v_maildir_delete(PyGObject *self)
{
	int ret;
	GError *error = NULL;

	ret = v_maildir_delete(V_MAILDIR(self->obj), &error);

	if (pyg_error_check(&error))
		return NULL;
	return PyBool_FromLong(ret);
}

static const PyMethodDef _PyVMaildir_methods[] = {
	{ "get_base", (PyCFunction)_wrap_v_maildir_get_base, METH_NOARGS, NULL },
	{ "delete", (PyCFunction)_wrap_v_maildir_delete, METH_NOARGS, NULL },
	{ NULL, NULL, 0, NULL }
};

PyTypeObject G_GNUC_INTERNAL PyVMaildir_Type = {
	PyObject_HEAD_INIT(NULL)
	0,						/* ob_size */
	"vtoaster.Maildir",				/* tp_name */
	sizeof(PyGObject),				/* tp_basicsize */
	0,						/* tp_itemsize */
	/* methods */
	(destructor)0,					/* tp_dealloc */
	(printfunc)0,					/* tp_print */
	(getattrfunc)0,					/* tp_getattr */
	(setattrfunc)0,					/* tp_setattr */
	(cmpfunc)0,					/* tp_compare */
	(reprfunc)0,					/* tp_repr */
	(PyNumberMethods*)NULL,				/* tp_as_number */
	(PySequenceMethods*)NULL,			/* tp_as_sequence */
	(PyMappingMethods*)NULL,			/* tp_as_mapping */
	(hashfunc)0,					/* tp_hash */
	(ternaryfunc)0,					/* tp_call */
	(reprfunc)0,					/* tp_str */
	(getattrofunc)0,				/* tp_getattro */
	(setattrofunc)0,				/* tp_setattro */
	(PyBufferProcs*)NULL,				/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	/* tp_flags */
	NULL,						/* Documentation string */
	(traverseproc)0,				/* tp_traverse */
	(inquiry)0,					/* tp_clear */
	(richcmpfunc)0,					/* tp_richcompare */
	offsetof(PyGObject, weakreflist),		/* tp_weaklistoffset */
	(getiterfunc)0,					/* tp_iter */
	(iternextfunc)0,				/* tp_iternext */
	(struct PyMethodDef*)_PyVMaildir_methods,	/* tp_methods */
	(struct PyMemberDef*)NULL,			/* tp_members */
	(struct PyGetSetDef*)NULL,			/* tp_getset */
	NULL,						/* tp_base */
	NULL,						/* tp_dict */
	(descrgetfunc)0,				/* tp_descr_get */
	(descrsetfunc)0,				/* tp_descr_set */
	offsetof(PyGObject, inst_dict),			/* tp_dictoffset */
	(initproc)_wrap_v_maildir_new,			/* tp_init */
	(allocfunc)0,					/* tp_alloc */
	(newfunc)0,					/* tp_new */
	(freefunc)0,					/* tp_free */
	(inquiry)0					/* tp_is_gc */
};



/* ----------- VAccount ----------- */

static int
_wrap_v_account_new(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "vdomain", "entry", NULL };
	PyGObject *vdomain, *entry;
	GError *error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!O!:VAccount.__init__", kwlist, &PyVDomain_Type, &vdomain, &PyVDomainEntry_Type, &entry))
		return -1;
	self->obj = (GObject *)v_account_new(V_DOMAIN(vdomain->obj), V_DOMAIN_ENTRY(entry->obj), &error);
	if (pyg_error_check(&error))
		return -1;

	if (!self->obj) {
		PyErr_SetString(PyExc_RuntimeError, "could not create VAccount object");
		return -1;
	}
	pygobject_register_wrapper((PyObject *)self);
	return 0;
}

static PyObject *
_wrap_v_account_get_maildir(PyGObject *self)
{
	VMaildir *maildir = v_account_get_maildir(V_ACCOUNT(self->obj));
	PyObject *ret = pygobject_new((GObject *)maildir);

	g_object_unref(maildir);
	return ret;
}

static PyObject *
_wrap_v_account_set_environent(PyGObject *self)
{
	v_account_set_environent(V_ACCOUNT(self->obj));

	Py_INCREF(Py_None);
	return Py_None;
}

static const PyMethodDef _PyVAccount_methods[] = {
	{ "get_maildir", (PyCFunction)_wrap_v_account_get_maildir, METH_NOARGS, NULL },
	{ "set_environent", (PyCFunction)_wrap_v_account_set_environent, METH_NOARGS, NULL },
	{ NULL, NULL, 0, NULL }
};

PyTypeObject G_GNUC_INTERNAL PyVAccount_Type = {
	PyObject_HEAD_INIT(NULL)
	0,						/* ob_size */
	"vtoaster.Account",				/* tp_name */
	sizeof(PyGObject),				/* tp_basicsize */
	0,						/* tp_itemsize */
	/* methods */
	(destructor)0,					/* tp_dealloc */
	(printfunc)0,					/* tp_print */
	(getattrfunc)0,					/* tp_getattr */
	(setattrfunc)0,					/* tp_setattr */
	(cmpfunc)0,					/* tp_compare */
	(reprfunc)0,					/* tp_repr */
	(PyNumberMethods*)NULL,				/* tp_as_number */
	(PySequenceMethods*)NULL,			/* tp_as_sequence */
	(PyMappingMethods*)NULL,			/* tp_as_mapping */
	(hashfunc)0,					/* tp_hash */
	(ternaryfunc)0,					/* tp_call */
	(reprfunc)0,					/* tp_str */
	(getattrofunc)0,				/* tp_getattro */
	(setattrofunc)0,				/* tp_setattro */
	(PyBufferProcs*)NULL,				/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	/* tp_flags */
	NULL,						/* Documentation string */
	(traverseproc)0,				/* tp_traverse */
	(inquiry)0,					/* tp_clear */
	(richcmpfunc)0,					/* tp_richcompare */
	offsetof(PyGObject, weakreflist),		/* tp_weaklistoffset */
	(getiterfunc)0,					/* tp_iter */
	(iternextfunc)0,				/* tp_iternext */
	(struct PyMethodDef*)_PyVAccount_methods,	/* tp_methods */
	(struct PyMemberDef*)NULL,			/* tp_members */
	(struct PyGetSetDef*)NULL,			/* tp_getset */
	NULL,						/* tp_base */
	NULL,						/* tp_dict */
	(descrgetfunc)0,				/* tp_descr_get */
	(descrsetfunc)0,				/* tp_descr_set */
	offsetof(PyGObject, inst_dict),			/* tp_dictoffset */
	(initproc)_wrap_v_account_new,			/* tp_init */
	(allocfunc)0,					/* tp_alloc */
	(newfunc)0,					/* tp_new */
	(freefunc)0,					/* tp_free */
	(inquiry)0					/* tp_is_gc */
};



/* ----------- VDomainIter ----------- */

typedef struct {
	PyObject_HEAD
	VDomainIter	*iter;
} PyVDomainIter;

static void
_PyVDomainIter_dealloc(PyVDomainIter *self)
{
	v_domain_iter_free(self->iter);
	PyObject_Del((PyObject*)self);
}

static PyObject *
_PyVDomainIter_iternext(PyVDomainIter *self)
{
	VDomainEntry *entry = v_domain_iter_next(self->iter, NULL);
	if (entry) {
		PyObject *ret = pygobject_new((GObject*)entry);

		g_object_unref(entry);
		return ret;
	} else {
		PyErr_SetNone(PyExc_StopIteration);
		return NULL;
	}
}

PyTypeObject G_GNUC_INTERNAL PyVDomainIter_Type = {
	PyObject_HEAD_INIT(NULL)
	0,						/* ob_size */
	"vtoaster.DomainIter",				/* tp_name */
	sizeof(PyVDomainIter),				/* tp_basicsize */
	0,						/* tp_itemsize */
	/* methods */
	(destructor)_PyVDomainIter_dealloc,		/* tp_dealloc */
	(printfunc)0,					/* tp_print */
	(getattrfunc)0,					/* tp_getattr */
	(setattrfunc)0,					/* tp_setattr */
	(cmpfunc)0,					/* tp_compare */
	(reprfunc)0,					/* tp_repr */
	(PyNumberMethods*)NULL,				/* tp_as_number */
	(PySequenceMethods*)NULL,			/* tp_as_sequence */
	(PyMappingMethods*)NULL,			/* tp_as_mapping */
	(hashfunc)0,					/* tp_hash */
	(ternaryfunc)0,					/* tp_call */
	(reprfunc)0,					/* tp_str */
	PyObject_GenericGetAttr,			/* tp_getattro */
	(setattrofunc)0,				/* tp_setattro */
	(PyBufferProcs*)NULL,				/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT,				/* tp_flags */
	"VDomain entry iterator",			/* Documentation string */
	(traverseproc)0,				/* tp_traverse */
	(inquiry)0,					/* tp_clear */
	(richcmpfunc)0,					/* tp_richcompare */
	0,						/* tp_weaklistoffset */
	PyObject_SelfIter,				/* tp_iter */
	(iternextfunc)_PyVDomainIter_iternext,		/* tp_iternext */
	(struct PyMethodDef*)NULL,			/* tp_methods */
	(struct PyMemberDef*)NULL,			/* tp_members */
	(struct PyGetSetDef*)NULL,			/* tp_getset */
	NULL,						/* tp_base */
	NULL,						/* tp_dict */
	(descrgetfunc)0,				/* tp_descr_get */
	(descrsetfunc)0,				/* tp_descr_set */
	0,						/* tp_dictoffset */
	(initproc)0,					/* tp_init */
	(allocfunc)0,					/* tp_alloc */
	(newfunc)0,					/* tp_new */
	(freefunc)0,					/* tp_free */
	(inquiry)0					/* tp_is_gc */
};

/* ----------- VDomain ----------- */

static int
_wrap_v_domain_new(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "type", "domain", "create", NULL };
	char *type = NULL, *domain = NULL;
	int create = 0;
	GError *error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|zzi:VDomain.__init__", kwlist, &type, &domain, &create))
		return -1;
	self->obj = (GObject *)v_domain_new(type, domain, create, &error);
	if (pyg_error_check(&error))
		return -1;

	if (!self->obj) {
		PyErr_SetString(PyExc_RuntimeError, "could not create VDomain object");
		return -1;
	}
	pygobject_register_wrapper((PyObject *)self);
	return 0;
}

static PyObject *
_wrap_v_domain_set_auth_method(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "type", NULL };
	PyObject *py_type = NULL;
	GType type;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O:VDomain.set_auth_method", kwlist, &py_type))
		return NULL;
	if ((type = pyg_type_from_object(py_type)) == 0)
		return NULL;

	v_domain_set_auth_method(V_DOMAIN(self->obj), type);

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
_wrap_v_domain_get_entry(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "account", NULL };
	char *account;
	GError *error = NULL;
	VDomainEntry *entry;
	PyObject *ret;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s:VDomain.get_entry", kwlist, &account))
		return NULL;

	entry = v_domain_get_entry(V_DOMAIN(self->obj), account, &error);

	if (pyg_error_check(&error))
		return NULL;

	ret = pygobject_new((GObject *)entry);
	g_object_unref(entry);
	return ret;
}

static PyObject *
_wrap_v_domain_get_account(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "account", NULL };
	char *account;
	GError *error = NULL;
	VAccount *ret;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s:VDomain.get_account", kwlist, &account))
		return NULL;

	ret = v_domain_get_account(V_DOMAIN(self->obj), account, &error);

	if (pyg_error_check(&error))
		return NULL;
	/* pygobject_new handles NULL checking */
	return pygobject_new((GObject *)ret);
}

static PyObject *
_wrap_v_domain_get_user(PyGObject *self)
{
	return pygobject_new((GObject *)v_domain_get_user(V_DOMAIN(self->obj)));
}

static PyObject *
_wrap_v_domain_auth(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "account", "password", NULL };
	char *account, *pass;
	int ret;
	GError *error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "ss:VDomain.auth", kwlist, &account, &pass))
		return NULL;

	ret = v_domain_auth(V_DOMAIN(self->obj), account, pass, &error);

	if (pyg_error_check(&error))
		return NULL;
	return PyBool_FromLong(ret);
}

static PyObject *
_wrap_v_domain_auth_account(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "account", "password", NULL };
	char *account, *pass;
	GError *error = NULL;
	VAccount *ret;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "ss:VDomain.auth_account", kwlist, &account, &pass))
		return NULL;

	ret = v_domain_auth_account(V_DOMAIN(self->obj), account, pass, &error);

	if (pyg_error_check(&error))
		return NULL;
	/* pygobject_new handles NULL checking */
	return pygobject_new((GObject *)ret);
}

static PyObject *
_wrap_v_domain_add_entry(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "entry", NULL };
	PyGObject *entry;
	int ret;
	GError *error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!:VDomain.add_entry", kwlist, &PyVDomainEntry_Type, &entry))
		return NULL;

	ret = v_domain_add_entry(V_DOMAIN(self->obj), V_DOMAIN_ENTRY(entry->obj), &error);

	if (pyg_error_check(&error))
		return NULL;
	return PyBool_FromLong(ret);
}

static PyObject *
_wrap_v_domain_modify_entry(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "entry", NULL };
	PyGObject *entry;
	int ret;
	GError *error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!:VDomain.modify_entry", kwlist, &PyVDomainEntry_Type, &entry))
		return NULL;

	ret = v_domain_modify_entry(V_DOMAIN(self->obj), V_DOMAIN_ENTRY(entry->obj), &error);

	if (pyg_error_check(&error))
		return NULL;
	return PyBool_FromLong(ret);
}

static PyObject *
_wrap_v_domain_remove_entry(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "account", NULL };
	char *account;
	int ret;
	GError *error = NULL;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s:VDomain.remove_entry", kwlist, &account))
		return NULL;

	ret = v_domain_remove_entry(V_DOMAIN(self->obj), account, &error);

	if (pyg_error_check(&error))
		return NULL;
	return PyBool_FromLong(ret);
}

static PyObject *
_wrap_v_domain_abort_changes(PyGObject *self)
{
	v_domain_abort_changes(V_DOMAIN(self->obj));

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
_wrap_v_domain_commit_changes(PyGObject *self)
{
	int ret;
	GError *error = NULL;

	ret = v_domain_commit_changes(V_DOMAIN(self->obj), &error);

	if (pyg_error_check(&error))
		return NULL;
	return PyBool_FromLong(ret);
}

static PyObject *
_wrap_v_domain_set_entry_password(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "entry", "pass", NULL };
	PyGObject *entry;
	char *pass;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!s:VDomain.set_entry_password", kwlist, &PyVDomainEntry_Type, &entry, &pass))
		return NULL;

	v_domain_set_entry_password(V_DOMAIN(self->obj), V_DOMAIN_ENTRY(entry->obj), pass);

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
_wrap_v_domain_build_userdir(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "name", NULL };
	char *name;
	gchar *ret;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s:VDomain.build_userdir", kwlist, &name))
		return NULL;

	ret = v_domain_build_userdir(V_DOMAIN(self->obj), name);

	if (ret) {
		PyObject *py_ret = PyString_FromString(ret);
		g_free(ret);
		return py_ret;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static const PyMethodDef _PyVDomain_methods[] = {
	{ "set_auth_method", (PyCFunction)_wrap_v_domain_set_auth_method, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "get_entry", (PyCFunction)_wrap_v_domain_get_entry, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "get_account", (PyCFunction)_wrap_v_domain_get_account, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "get_user", (PyCFunction)_wrap_v_domain_get_user, METH_NOARGS, NULL },
	{ "auth", (PyCFunction)_wrap_v_domain_auth, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "auth_account", (PyCFunction)_wrap_v_domain_auth_account, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "add_entry", (PyCFunction)_wrap_v_domain_add_entry, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "modify_entry", (PyCFunction)_wrap_v_domain_modify_entry, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "remove_entry", (PyCFunction)_wrap_v_domain_remove_entry, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "abort_changes", (PyCFunction)_wrap_v_domain_abort_changes, METH_NOARGS, NULL },
	{ "commit_changes", (PyCFunction)_wrap_v_domain_commit_changes, METH_NOARGS, NULL },
	{ "set_entry_password", (PyCFunction)_wrap_v_domain_set_entry_password, METH_VARARGS|METH_KEYWORDS, NULL },
	{ "build_userdir", (PyCFunction)_wrap_v_domain_build_userdir, METH_VARARGS|METH_KEYWORDS, NULL },
	{ NULL, NULL, 0, NULL }
};

static PyObject *
_PyVDomain_getiter(PyGObject *self)
{
	PyVDomainIter *ret;

	ret = PyObject_NEW(PyVDomainIter, &PyVDomainIter_Type);
	ret->iter = v_domain_iter_new(V_DOMAIN(self->obj));

	return (PyObject*)ret;
}

PyTypeObject G_GNUC_INTERNAL PyVDomain_Type = {
	PyObject_HEAD_INIT(NULL)
	0,						/* ob_size */
	"vtoaster.Domain",				/* tp_name */
	sizeof(PyGObject),				/* tp_basicsize */
	0,						/* tp_itemsize */
	/* methods */
	(destructor)0,					/* tp_dealloc */
	(printfunc)0,					/* tp_print */
	(getattrfunc)0,					/* tp_getattr */
	(setattrfunc)0,					/* tp_setattr */
	(cmpfunc)0,					/* tp_compare */
	(reprfunc)0,					/* tp_repr */
	(PyNumberMethods*)NULL,				/* tp_as_number */
	(PySequenceMethods*)NULL,			/* tp_as_sequence */
	(PyMappingMethods*)NULL,			/* tp_as_mapping */
	(hashfunc)0,					/* tp_hash */
	(ternaryfunc)0,					/* tp_call */
	(reprfunc)0,					/* tp_str */
	(getattrofunc)0,				/* tp_getattro */
	(setattrofunc)0,				/* tp_setattro */
	(PyBufferProcs*)NULL,				/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	/* tp_flags */
	NULL,						/* Documentation string */
	(traverseproc)0,				/* tp_traverse */
	(inquiry)0,					/* tp_clear */
	(richcmpfunc)0, 				/* tp_richcompare */
	offsetof(PyGObject, weakreflist),		/* tp_weaklistoffset */
	(getiterfunc)_PyVDomain_getiter,		/* tp_iter */
	(iternextfunc)0,				/* tp_iternext */
	(struct PyMethodDef*)_PyVDomain_methods,	/* tp_methods */
	(struct PyMemberDef*)NULL,			/* tp_members */
	(struct PyGetSetDef*)NULL,			/* tp_getset */
	NULL,						/* tp_base */
	NULL,						/* tp_dict */
	(descrgetfunc)0,				/* tp_descr_get */
	(descrsetfunc)0,				/* tp_descr_set */
	offsetof(PyGObject, inst_dict),			/* tp_dictoffset */
	(initproc)_wrap_v_domain_new,			/* tp_init */
	(allocfunc)0,					/* tp_alloc */
	(newfunc)0,					/* tp_new */
	(freefunc)0,					/* tp_free */
	(inquiry)0					/* tp_is_gc */
};



/* ----------- VDomainAuth ----------- */

PyTypeObject G_GNUC_INTERNAL PyVDomainAuth_Type = {
	PyObject_HEAD_INIT(NULL)
	0,						/* ob_size */
	"vtoaster.DomainAuth",				/* tp_name */
	sizeof(PyGObject),				/* tp_basicsize */
	0,						/* tp_itemsize */
	/* methods */
	(destructor)0,					/* tp_dealloc */
	(printfunc)0,					/* tp_print */
	(getattrfunc)0,					/* tp_getattr */
	(setattrfunc)0,					/* tp_setattr */
	(cmpfunc)0,					/* tp_compare */
	(reprfunc)0,					/* tp_repr */
	(PyNumberMethods*)NULL,				/* tp_as_number */
	(PySequenceMethods*)NULL,			/* tp_as_sequence */
	(PyMappingMethods*)NULL,			/* tp_as_mapping */
	(hashfunc)0,					/* tp_hash */
	(ternaryfunc)0,					/* tp_call */
	(reprfunc)0,					/* tp_str */
	(getattrofunc)0,				/* tp_getattro */
	(setattrofunc)0,				/* tp_setattro */
	(PyBufferProcs*)NULL,				/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	/* tp_flags */
	NULL,						/* Documentation string */
	(traverseproc)0,				/* tp_traverse */
	(inquiry)0,					/* tp_clear */
	(richcmpfunc)0,					/* tp_richcompare */
	offsetof(PyGObject, weakreflist),		/* tp_weaklistoffset */
	(getiterfunc)0,					/* tp_iter */
	(iternextfunc)0,				/* tp_iternext */
	(struct PyMethodDef*)NULL,			/* tp_methods */
	(struct PyMemberDef*)NULL,			/* tp_members */
	(struct PyGetSetDef*)NULL,			/* tp_getset */
	NULL,						/* tp_base */
	NULL,						/* tp_dict */
	(descrgetfunc)0,				/* tp_descr_get */
	(descrsetfunc)0,				/* tp_descr_set */
	offsetof(PyGObject, inst_dict),			/* tp_dictoffset */
	(initproc)0,					/* tp_init */
	(allocfunc)0,					/* tp_alloc */
	(newfunc)0,					/* tp_new */
	(freefunc)0,					/* tp_free */
	(inquiry)0					/* tp_is_gc */
};



/* ----------- VDomainEntry ----------- */

static int
_wrap_v_domain_entry_new(PyGObject *self, PyObject *args, PyObject *kwargs)
{
	static char *kwlist[] = { "name", "maildir", NULL };
	char *name, *maildir;

	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "ss:VDomainEntry.__init__", kwlist, &name, &maildir))
		return -1;
	self->obj = (GObject *)v_domain_entry_new(name, maildir);

	if (!self->obj) {
		PyErr_SetString(PyExc_RuntimeError, "could not create VDomainEntry object");
		return -1;
	}
	pygobject_register_wrapper((PyObject *)self);
	return 0;
}

static PyObject *
_wrap_v_domain_entry_copy(PyGObject *self)
{
	VDomainEntry *ret;

	ret = v_domain_entry_copy(V_DOMAIN_ENTRY(self->obj));

	/* pygobject_new handles NULL checking */
	return pygobject_new((GObject *)ret);
}

static PyObject *
_wrap_v_domain_entry_load_defaults(PyGObject *self)
{

	v_domain_entry_load_defaults(V_DOMAIN_ENTRY(self->obj));

	Py_INCREF(Py_None);
	return Py_None;
}

GETSET_MS(v_domain_entry, VDomainEntry, V_DOMAIN_ENTRY, name, name, 0)
GETSET_MS(v_domain_entry, VDomainEntry, V_DOMAIN_ENTRY, maildir, maildir, 1)
GETSET_MS(v_domain_entry, VDomainEntry, V_DOMAIN_ENTRY, passwd, password, 1)
GETSET_MU(v_domain_entry, VDomainEntry, V_DOMAIN_ENTRY, personal, personal, 1)

GETSET_X(v_domain_entry, V_DOMAIN_ENTRY, hardquota, Int, Long, Int)
GETSET_X(v_domain_entry, V_DOMAIN_ENTRY, softquota, Int, Long, Int)
GETSET_X(v_domain_entry, V_DOMAIN_ENTRY, msgsize, Int, Long, Int)
GETSET_X(v_domain_entry, V_DOMAIN_ENTRY, msgcount, Int, Long, Int)
GETSET_X(v_domain_entry, V_DOMAIN_ENTRY, ctime, Int, Long, Long)
GETSET_X(v_domain_entry, V_DOMAIN_ENTRY, expiry, Int, Long, Long)
GETSET_X(v_domain_entry, V_DOMAIN_ENTRY, enabled, Int, Long, Bool)

static PyObject*
_wrap_v_domain_entry_get_forwards(PyGObject *self, void *closure)
{
	PyObject *forwards;
	int len, pos;
	gchar **argv;

	argv = V_DOMAIN_ENTRY(self->obj)->forwards;
	if (!argv) {
		Py_INCREF(Py_None);
		return Py_None;
	}

	for(len = 0; argv[len]; len++);
	forwards = PyList_New(len);

	for(pos = 0; pos < len; pos++)
		PyList_SetItem(forwards, pos, PyString_FromString(argv[pos]));

	return forwards;
}

static int
_wrap_v_domain_entry_set_forwards(PyGObject *self, PyObject *value, void *closure)
{
	PyObject *arg;
	int len, pos;
	gchar **argv;

	if (!PyList_Check(value)) {
		PyErr_SetString(PyExc_TypeError,
			"VDomainEntry.set_forwards expects a list of strings.");
		return -1;
	}

	len = PyList_Size(value);
	if (len == -1) {
		PyErr_SetString(PyExc_TypeError,
			"VDomainEntry.set_forwards expects a list of strings.");
		return -1;
	}

	argv = g_new(gchar*, len + 1);
	argv[len] = NULL;
	for(pos = 0; pos < len; pos++) {
		arg = PyList_GetItem(value, pos);
		argv[pos] = g_strdup(PyString_AsString(arg));
		if (!argv[pos]) {
			g_strfreev(argv);
			return -1;
		}
	}

	v_domain_entry_set_forwards_v(V_DOMAIN_ENTRY(self->obj), argv);

	g_strfreev(argv);

	return 0;
}

static PyObject *
_wrap_v_domain_entry_dump(PyGObject *self)
{
	gchar *ret;

	ret = v_domain_entry_dump(V_DOMAIN_ENTRY(self->obj));

	if (ret) {
		PyObject *py_ret = PyString_FromString(ret);
		g_free(ret);
		return py_ret;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
_wrap_v_domain_entry_output(PyGObject *self)
{
	gchar *ret;

	ret = v_domain_entry_output(V_DOMAIN_ENTRY(self->obj));

	if (ret) {
		PyObject *py_ret = PyString_FromString(ret);
		g_free(ret);
		return py_ret;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static const PyMethodDef _PyVDomainEntry_methods[] = {
	{ "copy", (PyCFunction)_wrap_v_domain_entry_copy, METH_NOARGS, NULL },
	{ "load_defaults", (PyCFunction)_wrap_v_domain_entry_load_defaults, METH_NOARGS, NULL },
	{ "dump", (PyCFunction)_wrap_v_domain_entry_dump, METH_NOARGS, NULL },
	{ "output", (PyCFunction)_wrap_v_domain_entry_output, METH_NOARGS, NULL },
	{ NULL, NULL, 0, NULL }
};

static PyGetSetDef _PyVDomainEntry_getsets[] = {
	{ "__name__",	(getter)_wrap_v_domain_entry_get_name,		(setter)_wrap_v_domain_entry_set_name },
	{ "name",	(getter)_wrap_v_domain_entry_get_name,		(setter)_wrap_v_domain_entry_set_name },
	{ "maildir",	(getter)_wrap_v_domain_entry_get_maildir,	(setter)_wrap_v_domain_entry_set_maildir },
	{ "password",	(getter)_wrap_v_domain_entry_get_passwd,	(setter)_wrap_v_domain_entry_set_passwd },
	{ "forwards",	(getter)_wrap_v_domain_entry_get_forwards,	(setter)_wrap_v_domain_entry_set_forwards },
	{ "personal",	(getter)_wrap_v_domain_entry_get_personal,	(setter)_wrap_v_domain_entry_set_personal },
	{ "hardquota",	(getter)_wrap_v_domain_entry_get_hardquota,	(setter)_wrap_v_domain_entry_set_hardquota },
	{ "softquota",	(getter)_wrap_v_domain_entry_get_softquota,	(setter)_wrap_v_domain_entry_set_softquota },
	{ "msgsize",	(getter)_wrap_v_domain_entry_get_msgsize,	(setter)_wrap_v_domain_entry_set_msgsize },
	{ "msgcount",	(getter)_wrap_v_domain_entry_get_msgcount,	(setter)_wrap_v_domain_entry_set_msgcount },
	{ "ctime",	(getter)_wrap_v_domain_entry_get_ctime,		(setter)_wrap_v_domain_entry_set_ctime },
	{ "expiry",	(getter)_wrap_v_domain_entry_get_expiry,	(setter)_wrap_v_domain_entry_set_expiry },
	{ "enabled",	(getter)_wrap_v_domain_entry_get_enabled,	(setter)_wrap_v_domain_entry_set_enabled },
	{ NULL,		(getter)NULL,					(setter)NULL }
};

PyTypeObject G_GNUC_INTERNAL PyVDomainEntry_Type = {
	PyObject_HEAD_INIT(NULL)
	0,						/* ob_size */
	"vtoaster.Entry",				/* tp_name */
	sizeof(PyGObject),				/* tp_basicsize */
	0,						/* tp_itemsize */
	/* methods */
	(destructor)0,					/* tp_dealloc */
	(printfunc)0,					/* tp_print */
	(getattrfunc)0,					/* tp_getattr */
	(setattrfunc)0,					/* tp_setattr */
	(cmpfunc)0,					/* tp_compare */
	(reprfunc)0,					/* tp_repr */
	(PyNumberMethods*)NULL,				/* tp_as_number */
	(PySequenceMethods*)NULL,			/* tp_as_sequence */
	(PyMappingMethods*)NULL,			/* tp_as_mapping */
	(hashfunc)0,					/* tp_hash */
	(ternaryfunc)0,					/* tp_call */
	(reprfunc)0,					/* tp_str */
	(getattrofunc)0,				/* tp_getattro */
	(setattrofunc)0,				/* tp_setattro */
	(PyBufferProcs*)NULL,				/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	/* tp_flags */
	NULL,						/* Documentation string */
	(traverseproc)0,				/* tp_traverse */
	(inquiry)0,					/* tp_clear */
	(richcmpfunc)0,				/* tp_richcompare */
	offsetof(PyGObject, weakreflist),		/* tp_weaklistoffset */
	(getiterfunc)0,					/* tp_iter */
	(iternextfunc)0,				/* tp_iternext */
	(struct PyMethodDef*)_PyVDomainEntry_methods,	/* tp_methods */
	(struct PyMemberDef*)NULL,			/* tp_members */
	(struct PyGetSetDef*)_PyVDomainEntry_getsets,	/* tp_getset */
	NULL,						/* tp_base */
	NULL,						/* tp_dict */
	(descrgetfunc)0,				/* tp_descr_get */
	(descrsetfunc)0,				/* tp_descr_set */
	offsetof(PyGObject, inst_dict),			/* tp_dictoffset */
	(initproc)_wrap_v_domain_entry_new,		/* tp_init */
	(allocfunc)0,					/* tp_alloc */
	(newfunc)0,					/* tp_new */
	(freefunc)0,					/* tp_free */
	(inquiry)0					/* tp_is_gc */
};



/* ----------- Auth ----------- */

PyTypeObject G_GNUC_INTERNAL PyVAuth_Type = {
	PyObject_HEAD_INIT(NULL)
	0,						/* ob_size */
	"vtoaster.Auth",				/* tp_name */
	sizeof(PyGObject),				/* tp_basicsize */
	0,						/* tp_itemsize */
	/* methods */
	(destructor)0,					/* tp_dealloc */
	(printfunc)0,					/* tp_print */
	(getattrfunc)0,					/* tp_getattr */
	(setattrfunc)0,					/* tp_setattr */
	(cmpfunc)0,					/* tp_compare */
	(reprfunc)0,					/* tp_repr */
	(PyNumberMethods*)0,				/* tp_as_number */
	(PySequenceMethods*)0,				/* tp_as_sequence */
	(PyMappingMethods*)0,				/* tp_as_mapping */
	(hashfunc)0,					/* tp_hash */
	(ternaryfunc)0,					/* tp_call */
	(reprfunc)0,					/* tp_str */
	(getattrofunc)0,				/* tp_getattro */
	(setattrofunc)0,				/* tp_setattro */
	(PyBufferProcs*)0,				/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	/* tp_flags */
	NULL,						/* Documentation string */
	(traverseproc)0,				/* tp_traverse */
	(inquiry)0,					/* tp_clear */
	(richcmpfunc)0,					/* tp_richcompare */
	offsetof(PyGObject, weakreflist),		/* tp_weaklistoffset */
	(getiterfunc)0,					/* tp_iter */
	(iternextfunc)0,				/* tp_iternext */
	(struct PyMethodDef*)NULL,			/* tp_methods */
	(struct PyMemberDef*)0,				/* tp_members */
	(struct PyGetSetDef*)0,				/* tp_getset */
	NULL,						/* tp_base */
	NULL,						/* tp_dict */
	(descrgetfunc)0,				/* tp_descr_get */
	(descrsetfunc)0,				/* tp_descr_set */
	offsetof(PyGObject, inst_dict),			/* tp_dictoffset */
	(initproc)0,					/* tp_init */
	(allocfunc)0,					/* tp_alloc */
	(newfunc)0,					/* tp_new */
	(freefunc)0,					/* tp_free */
	(inquiry)0					/* tp_is_gc */
};

void initvtoaster(void);

DL_EXPORT(void) initvtoaster(void)
{
	PyObject *module, *dict;

	init_pygobject_check(2, 12, 0);
	g_assert(pygobject_register_class);

	if ((module = PyImport_ImportModule("gobject")) != NULL) {
		_PyGObject_Type = (PyTypeObject *)PyObject_GetAttrString(module, "GObject");
		if (!_PyGObject_Type) {
			PyErr_SetString(PyExc_ImportError,
				"cannot import name GObject from gobject");
			return;
		}
	} else {
		PyErr_SetString(PyExc_ImportError, "could not import gobject");
		return;
	}

	if ((module = PyImport_ImportModule("pwd")) != NULL) {
		_PyStruct_Passwd_Type = (PyTypeObject *)PyObject_GetAttrString(module, "struct_passwd");
		if (!_PyStruct_Passwd_Type) {
			PyErr_SetString(PyExc_ImportError,
				"cannot import name struct_passwd from pwd");
			return;
		}
	} else {
		PyErr_SetString(PyExc_ImportError, "could not import pwd");
		return;
	}

	module = Py_InitModule("vtoaster", NULL);
	dict = PyModule_GetDict(module);

	pygobject_register_class(dict, "Config", V_TYPE_CONFIG, &PyVConfig_Type, Py_BuildValue("(O)", &PyGObject_Type));
	pygobject_register_class(dict, "User", V_TYPE_USER, &PyVUser_Type, Py_BuildValue("(O)", &PyGObject_Type));
	pygobject_register_class(dict, "Virtual", V_TYPE_QMAIL_VIRTUAL, &PyVQMailVirtual_Type, Py_BuildValue("(O)", &PyGObject_Type));
	pygobject_register_class(dict, "Maildir", V_TYPE_MAILDIR, &PyVMaildir_Type, Py_BuildValue("(O)", &PyGObject_Type));
	pygobject_register_class(dict, "Account", V_TYPE_ACCOUNT, &PyVAccount_Type, Py_BuildValue("(O)", &PyGObject_Type));
	pygobject_register_class(dict, "Domain", V_TYPE_DOMAIN, &PyVDomain_Type, Py_BuildValue("(O)", &PyGObject_Type));
	pygobject_register_class(dict, "DomainAuth", V_TYPE_DOMAIN_AUTH, &PyVDomainAuth_Type, Py_BuildValue("(O)", &PyGObject_Type));
	pygobject_register_class(dict, "Entry", V_TYPE_DOMAIN_ENTRY, &PyVDomainEntry_Type, Py_BuildValue("(O)", &PyGObject_Type));
	pygobject_register_class(dict, "Auth", V_TYPE_AUTH, &PyVAuth_Type, Py_BuildValue("(O)", &PyVDomainAuth_Type));
	pyg_set_object_has_new_constructor(V_TYPE_AUTH);

	pyg_enum_add(module, "AuthError", "V_", V_TYPE_AUTH_ERROR);
	pyg_enum_add(module, "DomainError", "V_", V_TYPE_DOMAIN_ERROR);

	if (PyErr_Occurred())
		PyErr_Print();
}
