#!/usr/bin/python
# -*- coding: utf-8 -*-
import os, sys, time, base64, getopt
import PAM, SOAPpy
from Crypto.Hash import MD5
import gobject, vtoaster

# eigene Exception fuer VToaster-Fehler, etwas aufgeräumter
class VToasterError(Exception):
	def __init__(self, e):
		if e.domain == "v-domain-error-quark":
			self.type = 'domain'
			self.code = vtoaster.DomainError.__enum_values__[e.code].value_nick
		elif e.domain == "v-auth-error-quark":
			self.type = 'auth'
			self.code = vtoaster.AuthError.__enum_values__[e.code].value_nick
		else:
			self.type = 'unknown domain: ' + e.domain
			self.code = 'unknown code: ' +str(e.code)
		self.msg = e.message

class AuthError(Exception):
	def __init__(self, code, message):
		self.type = 'auth'
		self.code = code
		self.message = message

# verwandle VToaster.Entry in brauchbare Mailbox-Daten
def makeAccount(entry, domain):
	res = {}

	for id in ['name', 'maildir', 'forwards', 'personal', 'hardquota',
	           'softquota', 'msgsize', 'msgcount', 'ctime', 'expiry',
	           'enabled']:
		value = getattr(entry, id)
		if value != None:
			res[id] = value

	if res.has_key('maildir'):
		home = domain.get_user().home
		# /home/mail/bla_com/users/foo -> ./users/foo
		if res['maildir'][:len(home)] == home:
			res['maildir'] = '.' + res['maildir'][len(home):]

	return res

# PAM-Konversation (simuliert die interaktive Eingabe des Passworts)
def pamConv(auth, query_list, data):
	resp = []
	for query, type in query_list:
		if type == PAM.PAM_PROMPT_ECHO_ON:
			resp.append((raw_input(query), 0))
		elif type == PAM.PAM_PROMPT_ECHO_OFF:
			resp.append((data, 0))
		elif type == PAM.PAM_PROMPT_ERROR_MSG or \
		     type == PAM.PAM_PROMPT_TEXT_INFO:
			resp.append(('', 0))
		else:
			return None
	return resp

# Authentifikation per PAM
def pamAuth(user, password, service):
	auth = PAM.pam()
	auth.setUserData(password)
	auth.start(service)
	auth.set_item(PAM.PAM_USER, user)
	auth.set_item(PAM.PAM_CONV, pamConv)
	auth.authenticate()

# Zeit in 4 Byte quetschen
def encodeTime(seconds):
	now = long(seconds)
	return chr(now & 0xff) + \
	       chr((now >> 8) & 0xff) + \
	       chr((now >> 16) & 0xff) + \
	       chr((now >> 24) & 0xff)

# Zeit aus 4 Byte extrahieren
def decodeTime(str):
	return (ord(str[0]) & 0xff) + \
	       ((ord(str[1]) & 0xff) << 8) + \
	       ((ord(str[2]) & 0xff) << 16) + \
	       ((ord(str[3]) & 0xff) << 24)

# Authentifikationsschlüssel erstellen, aus Uhrzeit und Salt
def makeKey(salt):
	global secret
	hash = MD5.new()
	now = encodeTime(time.time())
	hash.update(secret)
	hash.update(now)
	hash.update(salt)
	hash.update(secret)
	return base64.b64encode(now + hash.digest())

# Authentifikationsschlüssel gegen Uhrzeit und Salt validieren
def validKey(key, salt):
	global timeout
	try:
		key = base64.b64decode(key)
	except:
		return False;
	if len(key) != 20:
		return False;

	now = time.time()
	start = decodeTime(key[:4])
	if now < start or now >= start + timeout:
		return False;

	hash = MD5.new()
	hash.update(secret)
	hash.update(key[:4])
	hash.update(salt)
	hash.update(secret)
	return key[4:] == hash.digest()

# Standardprozedere um aus dem Prozess einen Daemon zu machen
def daemonize():
	null_read = file('/dev/null', 'r' )
	null_write = file('/dev/null', 'a+' )

	try:
		os.dup2(null_read.fileno(), 0)
	except OSError, e:
		pass
	try:
		os.dup2(null_write.fileno(), 1)
	except OSError, e:
		pass
	try:
		os.dup2(null_write.fileno(), 2)
	except OSError, e:
		pass

	null_read.close()
	null_write.close()

	pid = os.fork()
	if pid > 0:
		os._exit(0)

	os.setsid()
	os.umask(077)

	pid = os.fork()
	if pid > 0:
		open('/var/run/vtoasterd.pid', 'w').write("%d\n"%pid)
		os._exit(0)

################
################ SOAP-Server-Funktionen
################

def loginDomain(domain, password):
	try:
		dom = vtoaster.Domain(None, domain)
		pamAuth(dom.get_user().user, password, 'pop')
		return makeKey(domain)

	except PAM.error, e:
		return AuthError('code %d' % e.args[1], e.args[0])
	except gobject.GError, e:
		return VToasterError(e)

def loginUser(user, domain, password):
	try:
		dom = vtoaster.Domain(None, domain)
		dom.set_auth_method(vtoaster.Auth)
		dom.auth_account(user, password)
		return makeKey(user + '@' + domain)

	except gobject.GError, e:
		return VToasterError(e)

def listDomain(key, domain):
	if not validKey(key, domain):
		return AuthError('timeout', 'Session timed out.')

	try:
		dom = vtoaster.Domain(None, domain)
		return map(lambda entry: makeAccount(entry, dom), dom)

	except gobject.GError, e:
		return VToasterError(e)

def listUser(key, user, domain):
	if not validKey(key, domain) and not validKey(key, user+'@'+domain):
		return AuthError('timeout', 'Session timed out.')

	try:
		dom = vtoaster.Domain(None, domain)
		return makeAccount(dom.get_entry(user), dom)

	except gobject.GError, e:
		return VToasterError(e)

################
################ Hauptprogramm
################

def usage():
	sys.stderr.write('Syntax: ' + sys.argv[0] + ' [OPTION]...\n')
	sys.stderr.write('\t-h, --help    show this help message\n')
	sys.stderr.write('\t-d, --debug   do not daemonize\n')


if __name__ == '__main__':
	try:
		opts, args = getopt.getopt(sys.argv[1:], "hd", \
				["help", "debug"])
	except getopt.GetoptError:
		usage()
		sys.exit(1)

	debug = False

	for o, a in opts:
		if o in ('-h', '--help'):
			usage()
			sys.exit()
		if o in ('-d', '--debug'):
			debug = True

	secret = base64.b64decode(open('/etc/vtoasterd.secret', 'r').readline())
	timeout = 3600

	server = SOAPpy.SOAPServer(("0.0.0.0", 323))

	if not debug:
		daemonize()

	server.registerFunction(loginDomain)
	server.registerFunction(loginUser)
	server.registerFunction(listDomain)
	server.registerFunction(listUser)

	try:
		server.serve_forever()
	except:
		pass
